#!/usr/bin/env python3
"""Tests for io."""

import os
from unittest.mock import patch

import nibabel as nib
import numpy as np
import pytest

from phys2cvr import io

# ## Unit tests


@pytest.mark.parametrize(
    'var, dtype, out',
    [
        (10, 'int', 10),
        (10.0, 'int', 10),
        ('10', 'int', 10),
        ([10], 'int', [10]),
        (None, 'int', None),
        (10, 'float', 10.0),
        (10.0, 'float', 10.0),
        ('10.0', 'float', 10.0),
        ([10.0], 'float', [10.0]),
        (None, 'float', None),
        (10, 'str', '10'),
        (10.0, 'str', '10.0'),
        ('10', 'str', '10'),
        ([10], 'str', '[10]'),
        (None, 'str', None),
        (10, 'list', [10]),
        ([10], 'list', [10]),
        ('10', 'list', ['10']),
        (None, 'list', None),
    ],
)
def test_if_declared_force_type(var, dtype, out):
    assert io.if_declared_force_type(var, dtype) == out


@patch('logging.warning')
@pytest.mark.parametrize(
    'var, dtype, varname, expected_warning',
    [
        (
            10,
            'int',
            'my_var',
            "Changing type of variable my_var from <class 'int'> to int",
        ),
        ('10', 'int', 'my_var', ''),
        (
            [10],
            'int',
            'my_var',
            "Changing type of variable my_var from <class 'list'> to int",
        ),
        (None, 'int', 'my_var', ''),
        (10.0, 'float', 'my_var', ''),
    ],
)
def test_if_declared_force_type_logging(mock_warning):
    io.if_declared_force_type(10, 'int', 'my_var')
    mock_warning.assert_called_with(
        "Changing type of variable my_var from <class 'int'> to int"
    )

    io.if_declared_force_type('10', 'int', 'my_var')
    mock_warning.assert_not_called()

    io.if_declared_force_type([10], 'int', 'my_var')
    mock_warning.assert_called_with(
        "Changing type of variable my_var from <class 'list'> to int"
    )

    io.if_declared_force_type(None, 'int', 'my_var')
    mock_warning.assert_not_called()

    io.if_declared_force_type(10.0, 'float', 'my_var', silent=True)
    mock_warning.assert_not_called()


def test_check_ext():
    all_ext = ['.csv', '.txt']
    fname = 'data.csv'
    assert io.check_ext(all_ext, fname) is True
    assert io.check_ext(all_ext, fname, remove=True) == ('data', True)
    all_ext = ['.CSV']
    assert io.check_ext(all_ext, fname) is True

    fname = 'data.xls'
    assert io.check_ext(all_ext, fname) is False
    assert io.check_ext(all_ext, fname, remove=True) == ('data.xls', False)

    all_ext = []
    fname = 'data.csv'
    assert io.check_ext(all_ext, fname, remove=True) == ('data.csv', False)


@patch('logging.warning')
def test_check_nifti_dim(mock_warning):
    fname = 'valid.nii.gz'
    data = np.ones((4, 4, 4, 4, 4))
    output = io.check_nifti_dim(fname, data, dim=5)
    assert output.ndim == 5

    output = io.check_nifti_dim(fname, data, dim=3)

    # Test if data has more dimensions than `dim`
    output = io.check_nifti_dim(fname, data, dim=3)
    mock_warning.assert_called_with(
        f'{fname} has more than 3 dimensions. Removing D > 3.'
    )
    assert output.ndim == 3


def test_load_nifti_get_mask(nifti_data, nifti_mask):
    # Test loading a nifti file with default arguments
    fname, expected_data = nifti_data
    data, mask, img = io.load_nifti_get_mask(fname)
    assert np.array_equal(data, expected_data)
    assert np.array_equal(mask, expected_data.any(axis=-1).squeeze())
    assert isinstance(img, nib.nifti1.Nifti1Image)

    # Test loading a nifti mask file with is_mask=True
    fname, expected_data = nifti_mask
    data, mask, img = io.load_nifti_get_mask(fname, is_mask=True)
    assert np.array_equal(data, expected_data)
    assert np.array_equal(mask, expected_data != 0)
    assert isinstance(img, nib.nifti1.Nifti1Image)


def test_export_regressor(testdir):
    petco2hrf_shift = np.random.rand(10)
    freq = 1
    tr = 2
    outname = os.path.join(testdir, 'test_regressor')
    suffix = 'petco2hrf'
    ext = '.1D'

    petco2hrf_demean = io.export_regressor(
        petco2hrf_shift, freq, tr, outname, suffix=suffix, ext=ext
    )

    pco2 = (petco2hrf_shift - petco2hrf_shift.mean())[::2]
    assert np.allclose(petco2hrf_demean, pco2)

    # Check if file was saved and has the correct content
    saved_file = np.loadtxt(f'{outname}_{suffix}{ext}')
    assert np.allclose(saved_file, pco2)


def test_export_nifti(testdir):
    # create some test data
    data = np.random.rand(10, 10, 10)
    affine = np.eye(4)
    header = nib.Nifti1Header()
    img = nib.Nifti1Image(data, affine, header)

    # create a temporary directory to store the output file
    fname = os.path.join(testdir, 'test_output.nii.gz')
    io.export_nifti(data, img, fname)

    # check if the output file exists
    assert os.path.exists(fname)

    # load the output file and check if the data matches the input
    out_img = nib.load(fname)
    assert out_img.get_fdata() == data
    assert out_img.affine == affine
    assert out_img.header == header


# ## Break tests


def test_break_if_declared_force_type():
    with pytest.raises(NotImplementedError) as errorinfo:
        io.if_declared_force_type('10', 'invalid_type')
    assert 'not supported' in str(errorinfo.value)


def test_break_check_nifti_dim_missing_dims():
    with pytest.raises(ValueError) as errorinfo:
        io.check_nifti_dim('missing_dims.nii.gz', np.ones((4, 4)), dim=4)
    assert 'seem to be a 4D file.' in str(errorinfo.value)



"""Hopefully importing everything."""

import pkgutil

from . import _version

__version__ = _version.get_versions()['version']

__all__ = []
for loader, module_name, _is_pkg in pkgutil.walk_packages(__path__):
    __all__.append(module_name)

    try:
        _module = loader.find_module(module_name).load_module(module_name)
    except AttributeError:
        _module = loader.find_spec(module_name).loader.load_module(module_name)
    globals()[module_name] = _module



#!/usr/bin/env python3
"""Tests for io."""

import os

import numpy as np

from phys2cvr import signal

# ## Unit tests


def test_spc():
    # Test case 1: test with a single time series
    ts = np.array([1, 2, 3, 4, 5])
    expected_result = np.array([-0.5, -0.25, 0, 0.25, 0.5])
    assert np.allclose(signal.spc(ts), expected_result)

    # Test case 2: test with multiple time series
    ts = np.array([[1, 2, 3, 4, 5], [5, 4, 3, 2, 1]])
    expected_result = np.array(
        [[-0.5, -0.25, 0, 0.25, 0.5], [0.5, 0.25, 0, -0.25, -0.5]]
    )
    assert np.allclose(signal.spc(ts), expected_result)

    # Test case 3: test with input ts having mean 0
    ts = np.array([[1, 2, 3], [-1, -2, -3]])
    expected_result = np.array([[1, 2, 3], [-1, -2, -3]])
    assert np.allclose(signal.spc(ts), expected_result)

    # Test case 4: test with input ts having nan values
    ts = np.array([[1, 2, np.nan], [-1, np.nan, -3]])
    expected_result = np.array([[1 / 3, 2 / 3, 0], [-1 / 2, 0, -1]])
    assert np.allclose(signal.spc(ts), expected_result)


def test_create_hrf():
    # Test case 1: test with default frequency
    expected_length = 528
    assert len(signal.create_hrf()) == expected_length

    # Test case 2: test with custom frequency
    freq = 100
    expected_length = 1057
    assert len(signal.create_hrf(freq)) == expected_length

    # Test case 3: test that the HRF is non-negative
    assert np.all(signal.create_hrf() >= 0)

    # Test case 4: test that the maximum value of HRF is 1
    assert signal.create_hrf().max() == 1

    # Test case 5: test that the sum of HRF is equal to 1
    assert np.isclose(signal.create_hrf().sum(), 1)

    # Test with custom frequency of 5
    freq = 5
    hrf = signal.create_hrf(freq)

    # Test that the length of the HRF is correct
    expected_length = 132
    assert len(hrf) == expected_length

    # Test that the HRF is non-negative
    assert np.all(hrf >= 0)

    # Test that the first peak of the HRF is at the correct time point
    expected_peak_time = 5
    actual_peak_time = np.argmax(hrf)
    assert actual_peak_time == expected_peak_time


def test_filter_signal():
    # Generate random data
    np.random.seed(0)
    data = np.random.rand(10, 20, 30)

    # Test with default filter parameters
    filtered_data = signal.filter_signal(data, tr=2)
    assert filtered_data.shape == (10, 20, 30)

    # Test with custom filter parameters
    filtered_data = signal.filter_signal(data, tr=2, lowcut=0.1, highcut=0.2)
    assert filtered_data.shape == (10, 20, 30)

    # Test that the filtered data is different from the original data
    assert not np.allclose(data, filtered_data)

    # Test that the filtered data has a lower variance than the original data
    assert np.var(filtered_data) < np.var(data)


# #!# Tolerance is too high, need to clean the files!
def test_convolve_petco2(testdir):
    # Create test data
    co2 = np.zeros(100)
    co2[20] = 1
    co2[40] = 1
    pidx = np.array([20, 40])
    freq = 10
    outname = str(os.path.join(testdir, 'test'))

    # Call function
    co2_conv = signal.convolve_petco2(co2, pidx, freq, outname)

    # Check output
    assert co2_conv.shape == co2.shape
    assert np.allclose(co2_conv[:19], 0)
    assert np.allclose(co2_conv[20], 1, rtol=1e-6)
    assert np.allclose(co2_conv[21:39], 0, rtol=1e-6)
    assert np.allclose(co2_conv[40], 1, rtol=1e-6)
    assert np.allclose(co2_conv[41:], 0)

    # Check files were created
    assert (os.path.join(testdir, 'test_petco2.png')).is_file()
    assert (os.path.join(testdir, 'test_petco2.1D')).is_file()
    assert (os.path.join(testdir, 'test_petco2hrf.png')).is_file()
    assert (os.path.join(testdir, 'test_petco2hrf.1D')).is_file()


def test_resample_signal():
    # Create a sinusoidal signal at 10 Hz
    t = np.linspace(0, 1, 1000)
    signal1 = np.sin(2 * np.pi * 10 * t)

    # Resample the signal to 5 Hz
    signal2 = signal.resample_signal(signal1, freq1=10, freq2=5)

    # Calculate expected signal at 5 Hz
    t2 = np.linspace(0, 1, 500)
    expected_signal2 = np.sin(2 * np.pi * 10 * t2)

    # Check if the resampled signal matches the expected signal
    assert np.allclose(signal2, expected_signal2, decimal=6)



#!/usr/bin/env python3
"""
Statistical module for phys2cvr.

Attributes
----------
LGR :
    Logger
"""

import logging
import os

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import zscore

from phys2cvr.io import FIGSIZE, SET_DPI, export_regressor
from phys2cvr.signal import resample_signal

R2MODEL = ['full', 'partial', 'intercept', 'adj_full', 'adj_partial', 'adj_intercept']

LGR = logging.getLogger(__name__)
LGR.setLevel(logging.INFO)


def x_corr(func, co2, n_shifts=None, offset=0, abs_xcorr=False):
    """
    Cross correlation between `func` and `co2`.

    Parameters
    ----------
    func : np.ndarray
        Timeseries, must be SHORTER (or of equal length) than `co2`
    co2 : np.ndarray
        Second timeseries, can be LONGER than `func`
    n_shifts : int or None, optional
        Number of shifts to consider when cross-correlating func and co2.
        When None (default), consider all possible shifts.
        Each shift consists of one sample.
    offset : int, optional
        Optional amount of offset desired for `func`, i.e. the amount of samples
        of `co2` to exclude from the cross correlation.
    abs_xcorr : bool, optional
        If True, x_corr will find the maximum absolute correlation,
        i.e. max(|corr(func, co2)|), rather than the maximum positive correlation.

    Returns
    -------
    float :
        Highest correlation
    int :
        Index of higher correlation
    xcorr : np.ndarray
        Full Xcorr

    Raises
    ------
    ValueError
        If `offset` is higher than the difference between the length of `co2` and `func`.
    NotImplementedError
        If `co2` length is smaller than `func` length.
    """
    if len(func) + offset > len(co2):
        if offset != 0:
            raise ValueError(
                f'The specified offset of {offset} is too high to '
                f'compare func of length {len(func)} with co2 of '
                f'length {len(co2)}'
            )
        else:
            raise NotImplementedError(
                f'The timeseries has length of {len(func)}, more than the'
                f'length of the given regressor ({len(co2)}). This case '
                'is not supported.'
            )

    if n_shifts is None:
        n_shifts = len(co2) - (len(func) + offset) + 1
        LGR.info(
            f'Considering all possible shifts of regressor for Xcorr, i.e. {n_shifts}'
        )
    else:
        if n_shifts + offset + len(func) > len(co2):
            LGR.warning(
                f'The specified amount of shifts ({n_shifts}) is too high for the '
                f'length of the regressor ({len(co2)}).'
            )
            n_shifts = len(co2) - (len(func) + offset) + 1
            LGR.warning(f'Considering {n_shifts} shifts instead.')

    xcorr = np.empty(n_shifts, dtype='float32')
    for n, i in enumerate(range(offset, n_shifts + offset)):
        xcorr[n] = (zscore(func) @ zscore(co2[0 + i : len(func) + i])) / len(func)

    if abs_xcorr:
        return np.abs(xcorr).max(), np.abs(xcorr).argmax() + offset, xcorr
    else:
        return xcorr.max(), xcorr.argmax() + offset, xcorr


def get_regr(
    func_avg,
    petco2hrf,
    tr,
    freq,
    outname,
    lag_max=None,
    trial_len=None,
    n_trials=None,
    ext='.1D',
    lagged_regression=True,
    legacy=False,
    abs_xcorr=False,
    skip_xcorr=False,
):
    """
    Create regressor(s) of interest for nifti GLM.

    Parameters
    ----------
    func_avg : np.ndarray
        Functional timeseries (1D)
    petco2hrf : np.ndarray
        Regressor of interest
    tr : str, int, or float
        TR of timeseries
    freq : str, int, or float
        Sample frequency of petco2hrf
    outname : list or path
        Path to output directory for regressors.
    lag_max : int or float, optional
        Limits (both positive and negative) of the temporal area to explore,
        expressed in seconds.
        Default: 9 (i.e. ±9 seconds)
    trial_len : str or int, optional
        Length of each single trial for tasks that have more than one
        (E.g. BreathHold, CO2 challenges, ...)
        Used to improve cross correlation estimation.
        Default: None
    n_trials : str or int, optional
        Number of trials in the task.
        Default: None
    ext : str, optional
        Extension to be used for the exported regressors.
    lagged_regression : bool, optional
        Estimate regressors for each possible lag of `petco2hrf`.
        If True, the maximum number of regressors will be `(freq*lag_max*2)+1`
    legacy : bool, optional
        If True, exclude the upper lag limit from the regression estimation.
        If True, the maximum number of regressors will be `(freq*lag_max*2)`
    abs_xcorr : bool, optional
        If True, the cross correlation will consider the maximum absolute
        correlation, i.e. if a negative correlation is higher than the highest
        positive, the negative correlation will be chosen instead.
    skip_xcorr : bool, optional
        If True, skip the cross correlation step.

    Returns
    -------
    petco2hrf_demean : np.ndarray
        The central, demeaned petco2hrf regressor.
    petco2hrf_shifts : np.ndarray
        The other shifted versions of the regressor.
    """
    # Setting up some variables
    first_tp, last_tp = 0, -1

    if trial_len and n_trials:
        # If both are specified, disregard two extreme _trial from matching.
        LGR.info(f'Specified {n_trials} trials lasting {trial_len} seconds')
        if n_trials > 2:
            LGR.info('Ignoring first trial to improve first bulk shift estimation')
            first_tp = int(trial_len * freq)
        else:
            LGR.info('Using all trials for bulk shift estimation')
        if n_trials > 3:
            LGR.info('Ignoring last trial to improve first bulk shift estimation')
            last_tp = first_tp * (n_trials - 1)

    elif trial_len and not n_trials:
        LGR.warning(
            'The length of trial was specified, but the number of '
            'trials was not. Using all trials for bulk shift estimation'
        )
    elif not trial_len and n_trials:
        LGR.warning(
            'The number of trials was specified, but the length of '
            'trial was not. Using all trials for bulk shift estimation'
        )
    else:
        LGR.info('Using all trials for bulk shift estimation.')

    # Upsample functional signal
    func_upsampled = resample_signal(func_avg, 1 / tr, freq)
    len_upd = func_upsampled.shape[0]

    # Preparing breathhold and CO2 trace for Xcorr
    func_cut = func_upsampled[first_tp:last_tp]
    petco2hrf_cut = petco2hrf[:last_tp]

    if not skip_xcorr:
        _, optshift, xcorr = x_corr(
            func_cut, petco2hrf_cut, n_shifts=None, offset=first_tp, abs_xcorr=abs_xcorr
        )
        LGR.info(f'Cross correlation estimated bulk shift at {optshift / freq} seconds')
        # Export estimated optimal shift in seconds
        with open(f'{outname}_optshift.1D', 'w') as f:
            print(f'{(optshift / freq):.4f}', file=f)

        # Preparing time axis for plots
        time_axis = np.linspace(0, (len(xcorr) - 1) / freq, len(xcorr))

        # Export xcorr figure
        plt.figure(figsize=FIGSIZE, dpi=SET_DPI)
        plt.plot(time_axis, xcorr)
        plt.plot(time_axis[optshift], xcorr[optshift], 'x')
        plt.legend(['Cross correlation value', 'Optimal detected shift'])
        plt.title('Cross correlation and optimal shift')
        plt.tight_layout()
        plt.savefig(f'{outname}_optshift.png', dpi=SET_DPI)
        plt.close()
    else:
        optshift = 0

    # This shouldn't happen, but still check
    if optshift + len_upd > len(petco2hrf):
        raise Exception(
            f'The found optimal shift {optshift / freq} removes too many samples to '
            'continue. This error should not be possible.'
        )

    petco2hrf_shift = petco2hrf[optshift : optshift + len_upd]

    # Exporting figures of shift
    plt.figure(figsize=FIGSIZE, dpi=SET_DPI)
    plt.plot(zscore(petco2hrf_shift), '-', zscore(func_upsampled), '-')
    plt.title('Optimally shifted regressor and average Grey Matter signal')
    plt.legend(['Optimally shifted regressor', 'Average Grey Matter signal'])
    plt.tight_layout()
    plt.savefig(f'{outname}_petco2hrf.png', dpi=SET_DPI)
    plt.close()

    petco2hrf_demean = export_regressor(
        petco2hrf_shift, freq, tr, outname, 'petco2hrf', ext
    )

    # Initialise the shifts first.
    petco2hrf_shifts = None

    if lagged_regression and lag_max:
        outprefix = os.path.join(
            os.path.split(outname)[0], 'regr', os.path.split(outname)[1]
        )
        os.makedirs(os.path.join(os.path.split(outname)[0], 'regr'), exist_ok=True)

        # Set num of fine shifts: 9 seconds is a bit more than physiologically feasible
        negrep = int(lag_max * freq)
        if legacy:
            posrep = negrep
        else:
            posrep = negrep + 1
        petco2hrf_shifts = np.empty(
            [func_avg.shape[0], negrep + posrep], dtype='float32'
        )

        # Padding regressor for shift, and padding optshift too

        if (optshift - negrep) < 0:
            lpad = negrep - optshift
        else:
            lpad = 0

        if (optshift + posrep + len_upd) > petco2hrf.shape[0]:
            rpad = (optshift + posrep + len_upd) - petco2hrf.shape[0]
        else:
            rpad = 0

        if func_cut.shape[0] <= petco2hrf.shape[0]:
            petco2hrf_padded = np.pad(petco2hrf, (int(lpad), int(rpad)), 'mean')
        elif func_cut.shape[0] > petco2hrf.shape[0]:
            petco2hrf_padded = np.pad(petco2hrf_shift, (int(lpad), int(rpad)), 'mean')

        for n, i in enumerate(range(-negrep, posrep)):
            petco2hrf_lagged = petco2hrf_padded[
                optshift + lpad - i : optshift + lpad - i + len_upd
            ]
            petco2hrf_shifts[:, n] = export_regressor(
                petco2hrf_lagged, freq, tr, outprefix, f'{n:04g}', ext
            )

    elif lagged_regression and lag_max is None:
        LGR.warning(
            'The generation of lagged regressors was requested, '
            'but the maximum lag was not specified. Skipping '
            'lagged regressor generation.'
        )
    else:
        LGR.info('Skipping lag regressors generation.')

    return petco2hrf_demean, petco2hrf_shifts


def get_legendre(degree, length):
    """
    Producesthe Legendre polynomials of order `degree`.

    Parameters
    ----------
    degree : int
        Highest order desired.
    length : int
        Number of samples of the polynomials.

    Returns
    -------
    legendre : np.ndarray
        A `degree`*`length` array with all the polynomials up to order `degree`
    """

    def _bonnet(d, x):
        if d == 0:
            return np.ones_like(x)
        elif d == 1:
            return x
        else:
            return (
                (2 * d - 1) * x * _bonnet(d - 1, x) - (d - 1) * _bonnet(d - 2, x)
            ) / d

    x = np.linspace(-1, 1, length)
    legendre = np.empty([length, degree + 1], dtype='float32')
    for n in range(degree + 1):
        legendre[:, n] = _bonnet(n, x)
    return legendre


def ols(Ymat, Xmat, r2model='full', residuals=False, demean=False):
    """
    Implement Ordinary Least Square linear regression.

    Both Ymat and Xmat must encode the time axis in axis 0.
    This is the barebone OLS implementation. For the full regression step,
    see `stats.regression`.

    Parameters
    ----------
    Ymat : np.ndarray
        Dependent variable, 1D or 2D.
        The time axis must be in columns, i.e. time axis must be axis 0.
    Xmat : np.ndarray
        Independent variables, 1D or 2D. The regressor of interest MUST be the last one.
        Regressors must be in columns, i.e. time axis must be axis 0.
    r2model : {'full', 'partial', 'intercept', 'adj_full', 'adj_partial', 'adj_intercept'}, optional
        R^2 to report.
        Possible models are:
            - 'full' (default)
                Use every regressor in the model, i.e. compare versus baseline 0
            - 'partial'
                Consider only the first vector of Xmat in the model, i.e. compare
                versus baseline composed by all vectors of Xmat beside the first.
            - 'intercept'
                Use every regressor in the model, but the intercept, i.e. compare
                versus baseline intercept (Legendre polynomial order 0, a.k.a.
                average signal)
            - 'adj_*'
                Adjusted R^2 version of simple counterpart
        Under normal conditions, while the R^2 value will be different between options,
        a lagged regression based on any R^2 model will give the same results.
        This WILL NOT be the case if orthogonalisations between the first vector of Xmat
        and the others are introduced: a lagged regression based on `partial` might hold
        different results from .
        Default: 'full'
    residuals : bool, optional
        If True, output only residuals of the model - this is mainly for orthogonalisation
        If False, output betas, tstats, R^2 (default).
    demean : bool, optional
        If True, demean Xmat before running OLS.
        Default is False.

    Returns
    -------
    betas : np.ndarray
        Beta values
    t_stats : np.ndarray
        T-stats values
    r_square : np.ndarray
        R^2 values

    Raises
    ------
    NotImplementedError
        If Ymat has more than 2 dimensions
        If Xmat has more than 2 dimensions
        At the moment, if "poly" R^2 is declared, as it's not implemented yet.
    ValueError
        If a non-valid R^2 value is declared
    """
    if Ymat.ndim > 2:
        raise NotImplementedError(
            'OLS on data with more than 2 dimensions is not implemented yet.'
        )
    elif Ymat.ndim < 2:
        Ymat = Ymat[..., np.newaxis]
    if Xmat.ndim > 2:
        raise NotImplementedError(
            'OLS with regressors with more than 2 dimensions is not implemented yet.'
        )
    elif Xmat.ndim < 2:
        Xmat = Xmat[..., np.newaxis]

    if demean:
        LGR.info('Demean regressors')
        Xmat = Xmat - Xmat.mean(axis=0)

    try:
        betas, RSS, _, _ = np.linalg.lstsq(Xmat, Ymat, rcond=None)
        if not RSS.any():
            RSS = np.zeros(betas.shape[1])

    except np.linalg.LinAlgError:
        raise ValueError(
            'The given matrices might not be oriented correctly. Try to transpose the '
            'regressor matrix.'
        )

    if residuals:
        return Ymat - (Xmat @ betas)
    else:
        # compute t-values of betas (estimates)
        # first compute number of degrees of freedom
        df = Xmat.shape[0] - Xmat.shape[1]

        # compute sigma:
        # RSS = sum{[mdata - (X * betas)]^2}
        # sigma = RSS / Degrees_of_Freedom
        # RSS = np.sum(np.power(Ymat.T - np.dot(Xmat, betas), 2), axis=0)
        sigma = RSS / df
        sigma = sigma[..., np.newaxis]

        # Copmute std of betas:
        # C = (mat^T * mat)_ii^(-1)
        # std(betas) = sqrt(sigma * C)
        C = np.diag(np.linalg.pinv(np.dot(Xmat.T, Xmat)))
        C = C[..., np.newaxis]
        std_betas = np.sqrt(np.outer(C, sigma))
        tstats = betas / std_betas
        tstats[np.isneginf(tstats)] = -9999
        tstats[np.isposinf(tstats)] = 9999

        # Compute R^2 coefficient of multiple determination!
        r2model_support = False
        for model in R2MODEL:
            if r2model == model:
                r2model_support = True
        if not r2model_support:
            raise ValueError(
                f'{r2model} R^2 not supported. Supported R^2 models are {R2MODEL}'
            )

        r2msg = ''
        # R^2 = 1 - RSS/TSS, (TSS = total sum of square)
        if 'full' in r2model:
            # Baseline model is 0 - this is what we're using ATM
            TSS = np.sum(np.power(Ymat, 2), axis=0)
        elif 'intercept' in r2model:
            # Baseline model is intercept: TSS is variance*samples
            TSS = Ymat.var(axis=0) * Ymat.shape[0]
        elif 'poly' in r2model:
            # Baseline model is legendre polynomials - or others: TSS is RSS of partial matrix
            # Could improve efficiency by moving this fitting step outside the regression loop
            # polynomials = Xmat[:, 0:4]
            # TSS = np.linalg.lstsq(polynomials, Ymat.T, rcond=None)[1]
            raise NotImplementedError("'poly' R^2 not implemented yet")
            r2msg = 'polynomial'
        elif 'partial' in r2model:
            pass

        if 'partial' in r2model:
            # We could also compute PARTIAL R square of regr instead (or on top)
            # See for computation: https://sscc.nimh.nih.gov/sscc/gangc/tr.html
            tstats_square = np.power(tstats, 2)
            r_square = (tstats_square / (tstats_square + df))[-1, :]
        else:
            r_square = np.ones(Ymat.shape[1], dtype='float32') - (RSS / TSS)

        if r2msg == '':
            r2msg = r2model
        if 'adj_' in r2model:
            # We could compute ADJUSTED R^2 instead
            r_square = 1 - ((1 - r_square) * (Xmat.shape[0] - 1) / (df - 1))
            r2msg = f'adjusted {r2msg}'

        LGR.info(f'Adopting {r2msg} baseline to compute R^2.')

        return betas, tstats, r_square


def regression(
    data,
    regr,
    denoise_mat=None,
    ortho_mat=None,
    extra_mat=None,
    mask=None,
    r2model='full',
    debug=False,
    x1D='mat.1D',
):
    """
    Estimate regression parameters.

    Parameters
    ----------
    data : np.ndarray
        Dependent variable of the model (i.e. Y), as a 4D volume.
    regr : np.ndarray
        Regressor of interest
    denoise_mat : np.ndarray or None, optional
        Confounding effects (regressors)
    ortho_mat : np.ndarray or None, optional
        Confounding effects (regressors) that will be orthogonalised w.r.t. `regr`, `denoise_mat`, and
        `extra_mat`
    extra_mat : np.ndarray or None, optional
        Extra factors (regressors) that will be used to orthogonalise `ortho_mat`
    mask : np.ndarray or None, optional
        A 3D mask to reduce the number of voxels to run the regression for.
    r2model : {'full', 'partial', 'intercept', 'adj_full', 'adj_partial', 'adj_intercept'}, optional
        R^2 to report.
        Possible models are:
            - 'full' (default)
                Use every regressor in the model, i.e. compare versus baseline 0
            - 'partial'
                Consider only `regr` in the model, i.e. compare versus baseline
                composed by all other regressors (`denoise_mat`)
            - 'intercept'
                Use every regressor in the model, but the intercept, i.e. compare
                versus baseline intercept (Legendre polynomial order 0, a.k.a.
                average signal)
            - 'adj_*'
                Adjusted R^2 version of normal counterpart
        Under normal conditions, while the R^2 value will be different between options,
        a lagged regression based on any R^2 model will give the same results.
        This WILL NOT be the case if orthogonalisations between `regr` and `denoise_mat`
        are introduced: a lagged regression based on `partial` might hold different
        results.
        Default: 'full'

    Returns
    -------
    bout : np.ndarray
        Beta map
    tout : np.ndarray
        T-stats map
    rout : np.ndarray
        R^2 map

    Raises
    ------
    ValueError
        If `denoise_mat` and `regr` do not have at least one common dimension.
    """
    # Obtain data in 2d
    if mask is None:
        mask = np.ones(data.shape[:-1])

    mask = mask.astype('bool')

    Ymat = data[mask]
    # Check that regr has "two" dimensions
    if regr.ndim < 2:
        regr = regr[..., np.newaxis]

    if denoise_mat is not None:
        if regr.shape[0] != denoise_mat.shape[0]:
            denoise_mat = denoise_mat.T
            if regr.shape[0] != denoise_mat.shape[0]:
                raise ValueError(
                    'The provided confounding matrix does not match '
                    'the dimensionality of the PetCO2hrf regressor!'
                )
        # Stack mat
        # Note: Xmat is not currently demeaned within this function, so inputs
        # should already be demeaned
        Xmat = np.hstack([denoise_mat, regr])
        if ortho_mat is not None:
            if ortho_mat.shape[0] != denoise_mat.shape[0]:
                ortho_mat = ortho_mat.T
                if ortho_mat.shape[0] != denoise_mat.shape[0]:
                    raise ValueError(
                        'The provided orthogonalised matrix does not match '
                        'the dimensionality of the PetCO2hrf regressor!'
                    )

            nuisance_mat = Xmat.copy()
            if extra_mat is not None:
                if extra_mat.shape[0] != denoise_mat.shape[0]:
                    extra_mat = extra_mat.T
                    if extra_mat.shape[0] != denoise_mat.shape[0]:
                        raise ValueError(
                            'The provided extra matrix does not match '
                            'the dimensionality of the PetCO2hrf regressor!'
                        )
                nuisance_mat = np.hstack([nuisance_mat, extra_mat])

            ortho_mat = ols(ortho_mat, nuisance_mat, residuals=True)
            Xmat = np.hstack([ortho_mat, Xmat])
    else:
        Xmat = regr

    if debug:
        os.makedirs(os.path.dirname(x1D), exist_ok=True)
        np.savetxt(x1D, Xmat, fmt='%.6f')

    betas, tstats, r_square = ols(
        Ymat.T, Xmat, r2model='full', residuals=False, demean=False
    )

    if debug:
        # debug should export betas, tstats, r_square
        pass

    # Assign betas, Rsquare and tstats to new volume
    bout = np.zeros(mask.shape)
    tout = np.zeros(mask.shape)
    rout = np.zeros(mask.shape)
    bout[mask] = betas[-1, :]
    tout[mask] = tstats[-1, :]
    rout[mask] = r_square
    return bout, tout, rout


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""



#!/usr/bin/env python3
"""Tests for io."""

import os

import numpy as np

from phys2cvr import stats

# ## Unit tests


def test_x_corr():
    # Define two signals, cosine and sine
    t = np.linspace(0, 1, 1000)
    func = np.cos(2 * np.pi * 5 * t)
    co2 = np.sin(2 * np.pi * 5 * t) + 0.1 * np.random.randn(len(t))

    # Calculate cross correlation
    corr, index, xcorr = stats.x_corr(func, co2, lastrep=100, offset=50)

    # Check that the highest correlation is close to 1
    assert corr > 0.95

    # Check that the index of the highest correlation is close to the true offset
    assert np.abs(index - 50) < 5

    # Check that the length of the cross correlation is correct
    assert len(xcorr) == 50

    # Check that the cross correlation is symmetric
    assert np.allclose(xcorr, xcorr[::-1])

    # Generate signals
    t = np.linspace(0, 10, 1000)
    func = np.sin(t)
    co2 = np.tile(np.cos(t), 2)
    co2 = np.tile(co2, 2)

    # Compute x_corr
    max_corr, max_index, xcorr = stats.x_corr(func, co2, abs_xcorr=True)

    # Compute expected result
    expected_max_corr = np.abs(np.corrcoef(func, co2[: len(func)].T)[1, 0])
    expected_max_index = np.argmax(np.abs(xcorr))

    # Check results
    assert np.isclose(max_corr, expected_max_corr, rtol=1e-6)
    assert max_index == expected_max_index



#!/usr/bin/env python3
"""
I/O and related utils.

Attributes
----------
EXT_1D : list
    List of supported TXT/1D file extensions, in lower case.
EXT_NIFTI : list
    List of supported nifti file extensions, in lower case.
FIGSIZE : tuple
    Figure size
SET_DPI : int
    DPI of the figure
LGR :
    Logger
"""

import logging
from os.path import exists
from pathlib import Path

import nibabel as nib
import numpy as np

from phys2cvr import signal

SET_DPI = 100
FIGSIZE = (18, 10)
EXT_1D = ['.txt', '.csv', '.tsv', '.1d', '.par', '.tsv.gz']
EXT_MAT = ['.mat']
# EXT_XLS = [".xls"]
EXT_ARRAY = EXT_1D + EXT_MAT  # + EXT_XLS
EXT_NIFTI = ['.nii', '.nii.gz']
EXT_GIFTI = ['.gii', '.gii.gz']
EXT_ALL = EXT_ARRAY + EXT_NIFTI + EXT_GIFTI


LGR = logging.getLogger(__name__)
LGR.setLevel(logging.INFO)


def if_declared_force_type(var, dtype, varname='an input variable', silent=False):
    """
    Make sure `var` is of type `dtype`.

    Parameters
    ----------
    var : str, int, or float
        Variable to change type of
    dtype : str
        Type to change `var` to
    varname : str, optional
        The name of the variable
    silent : bool, optional
        If True, don't return any message

    Returns
    -------
    int, float, str, list, or var
        The given `var` in the given `dtype`, or `var` if '' or None

    Raises
    ------
    NotImplementedError
        If dtype is not 'int', 'float', 'str', or 'list'
    """
    if var:
        if dtype == 'int':
            tmpvar = int(var)
        elif dtype == 'float':
            tmpvar = float(var)
        elif dtype == 'str':
            tmpvar = str(var)
        elif dtype == 'list':
            if type(var) is list:
                tmpvar = var
            else:
                tmpvar = [var]
        else:
            raise NotImplementedError(f'Type {dtype} not supported')

        if not silent:
            if type(tmpvar) is not type(var):
                if varname != 'an input variable':
                    varname = f'variable {varname}'

                LGR.warning(f'Changing type of {varname} from {type(var)} to {dtype}')

        return tmpvar

    else:
        return var


def check_ext(all_ext, fname, scan=False, remove=False):
    """Check which extension a file has, and possibly remove it.

    Parameters
    ----------
    all_ext : list
        All possible extensions to check within.
    fname : str or os.PathLike
        The filename to check.
    scan : bool, optional
        Scan the given path to see if there is a file with that extension
        If True and no path declared, check if fname has a path, if not scan '.'
        If False, don't scan any folder.
    remove : bool, optional
        Remove the extension from fname if it has one.

    Returns
    -------
    obj_return : Uses a list to return variable amount of options.
        has_ext : boolean
            True if the extension is found, false otherwise.
        fname : str or os.PathLike
            If ``remove`` is True, return (extensionless) fname.
        ext : str
            If both ``remove`` and ``has_ext`` are True, returns also found extension.
    """
    all_ext = if_declared_force_type(all_ext, 'list', silent=True)

    ext = ''.join(Path(fname).suffixes)

    LGR.debug(f'{fname} ends with extension {ext}')

    has_ext = True if ext in all_ext else False

    if not has_ext and scan:
        for ext in all_ext:
            if exists(f'{fname}{ext}'):
                fname = f'{fname}{ext}'
                LGR.warning(f'Found {fname}{ext}, using it as input henceforth')
                has_ext = True
                break

    obj_return = [has_ext]

    if remove:
        obj_return += [
            fname[: -len(ext)],
            None if ext == '' else ext,
        ]  # case insensitive solution
    else:
        obj_return += [fname]

    return obj_return[:]


def check_nifti_dim(fname, data, dim=4):
    """
    Remove extra dimensions.

    Parameters
    ----------
    fname : str
        The name of the file representing `data`
    data : np.ndarray
        The data which dimensionality needs to be checked
    dim : int, optional
        The amount of dimensions expected/desired in the data.

    Returns
    -------
    np.ndarray
        If `len(data.shape)` = `dim`, returns data.
        If `len(data.shape)` > `dim`, returns a version of data without the
        dimensions above `dim`.

    Raises
    ------
    ValueError
        If `data` has less dimensions than `dim`
    """
    if len(data.shape) < dim:
        raise ValueError(
            f'A {dim}D nifti file is required, but {fname} has {data.ndim}D. Please '
            'check the input file.'
        )
    if len(data.shape) > dim:
        LGR.warning(f'{fname} has more than {dim} dimensions. Removing D > {dim}.')
        for ax in range(dim, len(data.shape)):
            data = np.delete(data, np.s_[1:], axis=ax)

    return np.squeeze(data)


def check_array_dim(fname, data, shape=None):
    """Check dimensions of a matrix.

    For future 3D implementation, check MIPLabCH/nigsp's check_array_dim.

    Parameters
    ----------
    fname : str
        The name of the file representing ``data``.
    data : np.ndarray
        The data which dimensionality needs to be checked.
    shape : None | ``'square'`` | ``'rectangle'``
        Shape of matrix, if empty, skip shape check.

    Returns
    -------
    np.ndarray
        If ``data.ndim = 2``, returns data.
        If ``data.ndim = 1`` and ``shape == 'rectangle'``, returns data with added empty
        axis.

    Raises
    ------
    NotImplementedError
        If ``data`` has more than 2 dimensions.
    ValueError
        If ``data`` is empty
        If ``shape == 'square'`` and ``data`` dimensions have different lengths.
    """
    data = data.squeeze()
    LGR.info('Checking data shape.')

    if data.shape[0] == 0:
        raise ValueError(f'{fname} is empty!')
    if data.ndim > 2:
        raise NotImplementedError(
            f'Only matrices up to 2D are supported, but given matrix is {data.ndim}D.'
        )
    if shape is not None:
        if data.ndim == 1 and shape == 'rectangle':
            data = data[..., np.newaxis]
            LGR.warning(
                f'Rectangular matrix required, but {fname} is a vector. '
                'Adding empty dimension.'
            )
        if shape == 'square' and data.shape[0] != data.shape[1]:
            raise ValueError(
                f'Square matrix required, but {fname} matrix has shape {data.shape}.'
            )

    return data


def load_nifti_get_mask(fname, is_mask=False, dim=3):
    """
    Load a nifti-like file and returns its data, its image, and a 3d mask.

    Support all nibabel supported filetypes

    Parameters
    ----------
    fname : str
        The filename to read in
    is_mask : bool, optional
        If the file contains a mask.
        Default: False
    dim : int
        The number of dimensions expected in fname

    Returns
    -------
    data : np.ndarray
        Data from nifti file.
    mask : np.ndarray
        If `is_mask` is False, np.ndarray of one dimension less than data,
        in which any element that has at least a value different from zero
        in the last dimension of `data` is True.
        If `is_mask` is True, mask is a boolean representation of data.
    img : nib.img
        Image object from nibabel.
    """
    img = nib.load(fname)

    LGR.info(f'Loading {fname}')

    if check_ext(EXT_GIFTI, fname)[0]:
        data = img.agg_data().transpose()
    else:
        data = img.get_fdata()

    data = check_nifti_dim(fname, data, dim=dim)

    if is_mask:
        mask = data != 0
    else:
        mask = data.any(axis=-1).squeeze()

    return data, mask, img


def load_txt(fname, shape=None):
    """Read files in textual format.

    Parameters
    ----------
    fname : str | os.PathLike
        Path to the txt file.
    shape : None | ``'square'`` | ``'rectangle'``
        Shape of matrix, if empty, skip check.

    Returns
    -------
    mtx : numpy.ndarray
        Data matrix.

    See Also
    --------
    check_array_dim
    """
    LGR.info(f'Loading {fname}.')

    _, _, ext = check_ext(EXT_1D, fname, scan=True, remove=True)

    if ext in ['.csv', '.csv.gz']:
        delimiter = ','
    elif ext in ['.tsv', '.tsv.gz']:
        delimiter = '\t'
    elif ext in ['.txt', '.1d', '.par']:
        delimiter = ' '
    else:
        delimiter = None

    mtx = np.genfromtxt(fname, delimiter=delimiter)

    mtx = check_array_dim(fname, mtx, shape)

    return mtx


def load_mat(fname, shape=None):
    """Read files in MATLAB format.

    Assumes the existence of a matrix/vector in the mat file, rendered as
    a numpy.ndarray. If there is more than a matrix, the one with the largest
    size will be selected.

    Parameters
    ----------
    fname : str | os.PathLike
        Path to the ``.mat`` file.
    shape : None | ``'square'`` | ``'rectangle'``
        Shape of matrix, if empty, skip check.

    Returns
    -------
    mtx : numpy.ndarray
        Data matrix.

    Notes
    -----
    Requires module ``pymatreader`` to work.

    See Also
    --------
    check_array_dim

    Raises
    ------
    EOFError
        If the mat file does not contain matrix or vectors.
    ImportError
        If pymatreader is not installed or can't be read.
    """
    try:
        from pymatreader import read_mat
    except ImportError:
        raise ImportError(
            'pymatreader is required to import mat files. '
            'Please see install instructions.'
        )

    LGR.info(f'Loading {fname}.')
    data = read_mat(fname)

    data_keys = []
    for k in data.keys():
        # Check data key only if it's not hidden
        # (skip '__header__', '__version__', '__global__')
        if '__' not in k:
            LGR.info(
                f'Checking {fname} key {str(k)} content for data '
                '(float array/matrices in MATLAB).'
            )
            if type(data[k]) is np.ndarray:
                data_keys.append(k)

    if len(data_keys) < 1:
        raise EOFError(f'{fname} does not seem to contain a numeric matrix.')
    elif len(data_keys) > 1:
        LGR.warning(
            'Found multiple possible arrays to load. '
            'Selecting the biggest (highest pythonic size).'
        )

    key = data_keys[0]
    for k in data_keys[1:]:
        if data[k].size > data[key].size:
            key = k

    LGR.info(f'Selected data from MATLAB variable {key}')
    mtx = data[key]
    mtx = check_array_dim(fname, mtx, shape)

    return mtx


def load_xls(fname, shape=''):
    """Read files in xls format.

    Parameters
    ----------
    fname : str | os.PathLike
        Path to the xls file.
    shape : None | ``'square'`` | ``'rectangle'``
        Shape of matrix, if empty, skip check.

    See Also
    --------
    check_array_dim

    Raises
    ------
    NotImplementedError
        Spreadheet loading is not implemented yet.
    """
    raise NotImplementedError('Spreadsheet loading is not implemented yet')


def load_array(fname, shape=''):
    """Read files in text-like format.

    Parameters
    ----------
    fname : str | os.PathLike
        Path to txt-like file.
    shape : None | ``'square'`` | ``'rectangle'``
        Shape of matrix, if empty, skip check.

    See Also
    --------
    check_array_dim

    Raises
    ------
    NotImplementedError
        Spreadheet loading is not implemented yet.
    """
    _, _, ext = check_ext(EXT_ARRAY, fname, scan=True, remove=True)

    if ext in EXT_1D:
        mtx = load_txt(fname, shape=shape)
    elif ext in EXT_MAT:
        mtx = load_mat(fname, shape=shape)
    # elif ext in EXT_XLS:
    #     mtx = load_xls(fname, shape=shape)
    else:
        raise NotImplementedError(
            f'{fname} file extension {ext} was not found or is not supported yet'
        )

    return mtx


def export_regressor(petco2hrf_shift, freq, tr, outname, suffix='petco2hrf', ext='.1D'):
    """
    Export generated regressors for fMRI analysis.

    Parameters
    ----------
    petco2hrf_shift : np.ndarray
        The regressors that needs to be exported, in its original sample
    freq : int
        The frequency of the petco2hrf_shift (hence physiological data)
    tr : float
        The tr of the fMRI timeseries at which to export.
    outname : str or path
        Prefix of the output file - can contain a path.
    suffix : str, optional
        The suffix of the output file.
    ext : str, optional
        The extension of the output file.

    Returns
    -------
    petco2hrf_demean : np.ndarray
        Interpolated version of `petco2hrf_shift` in the sampling of the fMRI data.
    """
    petco2hrf_shift = signal.resample_signal(petco2hrf_shift, freq, 1 / tr)
    petco2hrf_demean = petco2hrf_shift - petco2hrf_shift.mean()
    np.savetxt(f'{outname}_{suffix}{ext}', petco2hrf_demean, fmt='%.6f')

    return petco2hrf_demean


def export_nifti(data, img, fname):
    """
    Export a nifti file.

    Parameters
    ----------
    data : np.ndarray
        Data to be exported
    img : nib.img
        Nibabel image object
    fname : str or path
        Name of the output file
    """
    klass = img.__class__

    out_img = klass(data, img.affine, img.header)
    out_img.to_filename(fname)


"""
Copyright 2021, Stefano Moia.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
